﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using MeshVR;
using static Oni;

/*
	SkyboxColor is plugin to change:
		Skybox/Panoramic
		Skybox/Cubemap
		Unlit/Texture (Skybox/6 Sided)
	material in CUA (.assetbundle)
    change their scale and compatible with portal system.

	Based on UnityAssetVamifier code from NoStage3
	Made by DemonLord
	License: Creative Commons with Attribution (CC BY)
*/

namespace DLPlugin {
	public class SkyboxColor : MVRScript {
		private readonly string VERSION_NUMBER = "1.1";

		//Valid shader names
		private readonly List<string> UNITY_SHADER_NAMES = new List<string>(new string[] {
            "Custom/StencilSkybox/Cubemap",
            "Custom/StencilSkybox/Panoramic",
            "Custom/StencilSkybox/6 Sided",
            "Skybox/Panoramic",
			"Skybox/Cubemap",
			"Unlit/Texture",
			"Skybox/6 Sided"
		});
		//material names for "Skybox/6 Sided" case
		private readonly List<string> SIDE_NAMES = new List<string>(new string[] {
			"back (instance)",
			"bottom (instance)",
			"left (instance)",
			"right (instance)",
			"front (instance)",
			"top (instance)",
			"up (instance)",
			"down (instance)"
		});


        protected JSONStorableFloat jsScaleX;
        protected JSONStorableFloat jsScaleY;
        protected JSONStorableFloat jsScaleZ;
        protected JSONStorableFloat jExposureFloat;
		protected JSONStorableColor jTintColor;
		protected JSONStorableBool showMessagesBool;
        protected JSONStorableBool r1Enable;
        protected JSONStorableStringChooser jpopStencilRef;

        protected List<MatPackage> origMaterials = new List<MatPackage>();
		protected List<MatPackage> newMaterials = new List<MatPackage>();
		protected List<Renderer> allRenderers = new List<Renderer>();
		protected CustomUnityAssetLoader cUAL;
        protected CustomUnityAssetLoader cUAL2;
        protected ReflectionProbe r1;
        protected List<Renderer> allRenderers2 = new List<Renderer>();
        protected Atom shadersAtom;
        protected List<string> layers32 = new List<string>();
        protected bool assetChanged;
        private Transform rescaleObject;
        private Vector3 scale;
        private Vector3 scaleWas;
        //private Atom selfAtom;

        public struct MatPackage {
			public int rendererIndex;
			public int materialIndex;
			public Material material;
			public bool isConverted;

			public MatPackage(int rIndex, int mIndex, Material mat, bool converted) {
				this.rendererIndex = rIndex;
				this.materialIndex = mIndex;
				this.material = mat;
				this.isConverted = converted;
			}
		}

		public override void Init() {
			try
            {
                //selfAtom = SuperController.singleton.GetAtomByUid(containingAtom.name);
                //CustomUnityAssetLoader c = containingAtom.childAtomContainer.GetComponentInChildren<CustomUnityAssetLoader>();
                //if (c != null) SuperController.LogMessage("CustomUnityAssetLoader "+c.transform.name);
                //ShowName(selfAtom.transform, "");
                for (int i = 0; i < 32; i++)
                {
                    layers32.Add(i.ToString());
                }

                for (int i = 0; i < containingAtom.childAtomContainer.childCount; i++)
                {
                    //SuperController.LogMessage("GetChild" + containingAtom.childAtomContainer.GetChild(i).name);
                    if (containingAtom.childAtomContainer.GetChild(i).name == "rescaleObject")
                    {
                        rescaleObject = containingAtom.childAtomContainer.GetChild(i);
                        scale = new Vector3(rescaleObject.transform.localScale.x, rescaleObject.transform.localScale.y, rescaleObject.transform.localScale.z);
                        scaleWas = scale;
                    }
                }

                // build/load/populate material controls for VAM UI
                BuildUIControls();

                AddVersionNumberToUI();
                foreach (Atom a in SuperController.singleton.GetAtoms())
				{
					if (a.name.ToLower() == "skyboxshaders")
                    {
                        shadersAtom = a;
						//SuperController.LogMessage("find shadersAtom");
						cUAL2 = shadersAtom.childAtomContainer.GetComponentInChildren<CustomUnityAssetLoader>();
                        cUAL2.RegisterAssetLoadedCallback(AssetLoaded2);
                        cUAL2.RegisterAssetClearedCallback(AssetCleared2);
                        if (cUAL2.isAssetLoaded)
                        {
                            AssetLoaded2();
                        }
                        CreateScrollablePopup(jpopStencilRef);
                    }
                }
				if (shadersAtom == null)
                {
                    cUAL = containingAtom.childAtomContainer.GetComponentInChildren<CustomUnityAssetLoader>();
                    //SuperController.LogMessage("find shadersAtom"+(cUAL==null));
                    // get the loader component
                    SetCUACallback(cUAL);
                }
                //SuperController.LogMessage("1 allRenderers: " + allRenderers.Count);
				//SuperController.LogMessage("1 allRenderers2: " + allRenderers2.Count);
			}
			catch (Exception e)
			{
				SuperController.LogError("SkyboxColor: " + e);
			}
		}
		
        protected void SetCUACallback(CustomUnityAssetLoader cUAL)
        {
            if (cUAL != null)
            {
                cUAL.RegisterAssetClearedCallback(AssetCleared);
                cUAL.RegisterAssetLoadedCallback(AssetLoaded);
                if (cUAL.isAssetLoaded)
                {
                    AssetLoaded();
                }
            }
        }
        
        /*
        protected void ShowName(Transform t, string tabs)
        {
            tabs += "  ";
            SuperController.LogMessage(tabs + "name : " + t.name*//* + " pos: " + t.transform.position*//*);
            //SuperController.LogMessage("Plane " + t.transform.position);
            for (int i = 0; i < t.childCount; i++)
            {
                ShowName(t.GetChild(i), tabs);
            }
        }*/



        protected void BuildUIControls() {
			// Exposure 
			jExposureFloat = new JSONStorableFloat("Exposure", 0.5f, SetExposure, 0f, 5f, true);
			RegisterFloat(jExposureFloat);
            CreateSlider(jExposureFloat);

			//Tint Color
			HSVColor tintColorHSVC = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
			jTintColor = new JSONStorableColor("TintColor", tintColorHSVC, SetTintColor);
			RegisterColor(jTintColor);
            CreateColorPicker(jTintColor, true);

            // r1EnableToggle
            r1Enable = new JSONStorableBool("Reflectionprobe", true, Setr1);
            RegisterBool(r1Enable);

            // ScrollablePopupStencilRef
            jpopStencilRef = new JSONStorableStringChooser("StencilRef", new List<string>(layers32), "0", "StencilRef", SetpopStencilRef);
            RegisterStringChooser(jpopStencilRef);
            jsScaleX = new JSONStorableFloat("scaleX", rescaleObject.transform.localScale.x, SetScaleXYZ, 0f, 10f, false);
            RegisterFloat(jsScaleX);
            CreateSlider(jsScaleX);
            jsScaleY = new JSONStorableFloat("scaleY", rescaleObject.transform.localScale.y, SetScaleXYZ, 0f, 10f, false);
            RegisterFloat(jsScaleY);
            CreateSlider(jsScaleY);
            jsScaleZ = new JSONStorableFloat("scaleZ", rescaleObject.transform.localScale.z, SetScaleXYZ, 0f, 10f, false);
            RegisterFloat(jsScaleZ);
            CreateSlider(jsScaleZ);

            // showMessagesToggle
            showMessagesBool = new JSONStorableBool("Show Messages", true);
			RegisterBool(showMessagesBool);
            CreateToggle(showMessagesBool, true);
		}


		protected void AddVersionNumberToUI() {
			JSONStorableString versionString = new JSONStorableString("pluginVersion", "\nSkyboxColor "+ VERSION_NUMBER + "\nby DemonLord (CC BY)");
			UIDynamicTextField textFieldVersion = CreateTextField(versionString, true);
			textFieldVersion.height = 120;
		}


		// determine if the material should be converted or not
		protected bool IsValidMaterial(Material mat) {
			if (!UNITY_SHADER_NAMES.Contains(mat.shader.name))
				return false;
			return true;
		}
		
		// check material name as one of 6 Sided names
		protected bool IsValidSide(Material mat) {
			if (!SIDE_NAMES.Contains(mat.name.ToLower()))
				return false;
			return true;
		}

		protected void AssetCleared() {
			// don't need to reset materials because the asset would have been destroyed
			allRenderers.Clear();
            origMaterials.Clear();
			newMaterials.Clear();
			RemoveToggle(r1Enable);
        }
        protected void AssetCleared2()
        {
            // don't need to reset materials because the asset would have been destroyed
            RestoreOriginalMaterials();
            AssetCleared();
            allRenderers2.Clear();
			AssetLoaded();
        }

        protected void AssetLoaded()
        {
            containingAtom.childAtomContainer.GetComponentsInChildren<Renderer>(false, allRenderers);
            ConvertMaterial();
        }
        protected void AssetLoaded2()
        {
            shadersAtom.childAtomContainer.GetComponentsInChildren<Renderer>(false, allRenderers2);
            cUAL = containingAtom.childAtomContainer.GetComponentInChildren<CustomUnityAssetLoader>();
            // get the loader component
            SetCUACallback(cUAL);
        }

        // wait for the asset bundle to be loaded by checking for renderers every 0.5s (a bit hacky...)
        protected IEnumerator WaitForAssetBundle() {

			while (allRenderers.Count < 1) {
				containingAtom.childAtomContainer.GetComponentsInChildren<Renderer>(false, allRenderers);
				yield return new WaitForSeconds(0.5f);
			}

			ConvertMaterial();
		}


		protected void ConvertMaterial() {
			List<Material> invalidMaterials = new List<Material>();

			if (allRenderers.Count == 0) {
				if (showMessagesBool.val) {
					SuperController.LogMessage("SkyboxColor: Make sure you have an .assetbundle loaded and a prop selected from the Asset dropdown. Then reload this plugin.");
				}
				return;
			}
            r1 = containingAtom.childAtomContainer.GetComponentInChildren<ReflectionProbe>(true);
            if (r1 != null)
            {
                CreateToggle(r1Enable, false);
            }
            // fill newMaterials with copy of the original material or an new material with the Skybox/6 Sided shader
            for (int i = 0; i < allRenderers.Count; i++) {
				for (int j = 0; j < allRenderers[i].materials.Length; j++) {
					Material mat = allRenderers[i].materials[j];
					origMaterials.Add(new MatPackage(i, j, mat, false));

					if (IsValidMaterial(mat))
                    {
                        if (mat.shader.name.Contains("Unlit/Texture") && IsValidSide(mat))
                        {
                            if (shadersAtom != null && allRenderers2.Count != 0)
                            {
								for (int i2 = 0; i2 < allRenderers2.Count; i2++)
								{
									if (allRenderers2[i2].material.shader.name.Contains("Skybox/6 Sided"))
									{
										newMaterials.Add(new MatPackage(i, j, new Material(allRenderers2[i2].material.shader), true));
										break;
									}
                                }
                            }
                            else
                            {
                                newMaterials.Add(new MatPackage(i, j, new Material(Shader.Find("Skybox/6 Sided")), true));
                            }
                        }
                        else
                        {
                            if (shadersAtom != null && allRenderers2.Count != 0)
                            {
								for (int i2 = 0; i2 < allRenderers2.Count; i2++)
								{
									if (allRenderers2[i2].material.shader.name.Contains(mat.shader.name)) 
									{
										newMaterials.Add(new MatPackage(i, j, new Material(allRenderers2[i2].material.shader), true));
										break;
									}
                                }
                            }
                            else
                            {
                                newMaterials.Add(new MatPackage(i, j, mat, true));
                            }
                        }
					} else {
						newMaterials.Add(new MatPackage(i, j, mat, false));
						invalidMaterials.Add(mat);
					}
				}
			}


			// output invalid/skipped materials
			foreach (Material invalidMaterial in invalidMaterials)
			{
				if (showMessagesBool.val)
				{
					SuperController.LogMessage("SkyboxColor: Material skipped: " + invalidMaterial.name + " -- Shader: [" + invalidMaterial.shader.name + "]");
				}
			}

			if (newMaterials.Count == 0) {
				SuperController.LogError("SkyboxColor: No compatible material could be found on this asset. Check message log for skipped materials/shaders.");
				return;
			}


			// replace renders from original to copy materials
			for (int i = 0; i < allRenderers.Count; i++) {
				Material[] matsToAssign = allRenderers[i].materials;
				for (int j = 0; j < allRenderers[i].materials.Length; j++) {
					Material mat = allRenderers[i].materials[j];
					if (IsValidMaterial(mat)) {
						Material vamMat = GetVamMaterialByIndex(i, j);
                        Material origMat = GetOrigMaterialByIndex(i, j);
                        //vamMat.CopyPropertiesFromMaterial(origMat);
                        // "Skybox/6 Sided" case
                        if (mat.shader.name.Contains("Unlit/Texture") && IsValidSide(mat)){
							vamMat.SetTexture("_LeftTex", origMat.GetTexture("_MainTex"));
							vamMat.SetTexture("_RightTex", origMat.GetTexture("_MainTex"));
							vamMat.SetTexture("_FrontTex", origMat.GetTexture("_MainTex"));
							vamMat.SetTexture("_BackTex", origMat.GetTexture("_MainTex"));
							vamMat.SetTexture("_UpTex", origMat.GetTexture("_MainTex"));
							vamMat.SetTexture("_DownTex", origMat.GetTexture("_MainTex"));
						}
                        if (mat.shader.name.Contains("Skybox/Cubemap"))
                        {
                            vamMat.SetTexture("_Tex", origMat.GetTexture("_Tex"));
                        }
                        if (mat.shader.name.Contains("Skybox/Panoramic"))
                        {
                            vamMat.SetTexture("_MainTex", origMat.GetTexture("_MainTex"));
                            vamMat.SetFloat("_Mapping", origMat.GetFloat("_Mapping"));
                            vamMat.SetFloat("_ImageType", origMat.GetFloat("_ImageType"));
                            vamMat.SetFloat("_MirrorOnBack", origMat.GetFloat("_MirrorOnBack"));
                            vamMat.SetFloat("_Layout", origMat.GetFloat("_Layout"));
                        }
                        matsToAssign[j] = vamMat;
					}
				}
				// assign new materials to prop renderer as array
				allRenderers[i].materials = matsToAssign;
            }
            //SuperController.LogMessage("origMaterials: " + origMaterials.Count);
            //SuperController.LogMessage("newMaterials: " + newMaterials.Count);

            JSONNode node;
            if (GetStoredNode("StencilRef", out node))
            {
                //if (node == null) continue;
                jpopStencilRef.valNoCallback = node.Value;
            }
            // apply user definable values initially
            SetExposure(jExposureFloat);
			SetTintColor(jTintColor);
			SetpopStencilRef(jpopStencilRef.val);
            assetChanged = true;
        }
        public bool GetStoredNode(string storableName, out JSONNode node)
        {
            node = null;
            if (assetChanged || (SuperController.singleton.loadJson?["atoms"] ?? null) == null) return false;
            foreach (JSONNode a in SuperController.singleton.loadJson["atoms"].AsArray)
            {
                if (a["id"].Value == containingAtom.name)
                {
                    foreach (JSONNode n in a["storables"].AsArray)
                    {
                        if (n["id"].Value == this.name)
                        {
                            if (n[storableName] != null)
                            {
                                node = n[storableName];
                                return true;
                            }
                        }
                    }
                }
            }
            return false;
        }

        protected void SetScaleXYZ(JSONStorableFloat jf)
        {
            if (jf.name == "scaleX") scale.x = jf.val;
            if (jf.name == "scaleY") scale.y = jf.val;
            if (jf.name == "scaleZ") scale.z = jf.val;
            rescaleObject.transform.localScale = scale;
            //SuperController.LogMessage("scale "+x1+" "+y1+" "+z1);
        }
        protected void Setr1(bool b)
        {
            if (r1 != null) r1.enabled = b;
        }

        protected void SetpopStencilRef(string v)
        {
            if (allRenderers.Count > 0)
            {
                foreach (MatPackage matPackage in newMaterials)
                {
                    if (!matPackage.isConverted) continue;
                    if (matPackage.material.HasProperty("_StencilRef"))
                        matPackage.material.SetInt("_StencilRef", Int32.Parse(v));
                }
            }
            
        }
        protected Material GetVamMaterialByIndex(int rIndex, int mIndex) {
			foreach (MatPackage matPackage in newMaterials) {
				if (matPackage.rendererIndex == rIndex && matPackage.materialIndex == mIndex) {
					return matPackage.material;
				}
			}
			return null;
		}

		void OnDestroy()
        {
            rescaleObject.transform.localScale = scaleWas;
            if (r1 != null) r1.enabled = true;
            RestoreOriginalMaterials();

            if (cUAL != null)
            {
                cUAL.DeregisterAssetClearedCallback(AssetCleared);
                cUAL.DeregisterAssetLoadedCallback(AssetLoaded);
            }
            if (cUAL2 != null)
            {
                cUAL2.DeregisterAssetClearedCallback(AssetCleared2);
                cUAL2.DeregisterAssetLoadedCallback(AssetLoaded2);
            }
        }

		protected Material GetOrigMaterialByIndex(int rIndex, int mIndex) {
			foreach (MatPackage matPackage in origMaterials) {
				if (matPackage.rendererIndex == rIndex && matPackage.materialIndex == mIndex) {
					return matPackage.material;
				}
			}
			return null;
		}

		protected void RestoreOriginalMaterials() {
			for (int i = 0; i < allRenderers.Count; i++) {
				if (allRenderers[i] != null) {   // renderer may have been destroyed already by atom remove
					List<Material> matsToAssign = new List<Material>();
					for (int j = 0; j < allRenderers[i].materials.Length; j++) {
						matsToAssign.Add(GetOrigMaterialByIndex(i, j));
					}
					allRenderers[i].materials = matsToAssign.ToArray();
				}
			}
		}

		protected void SetExposure(JSONStorableFloat jf) {
			foreach (MatPackage matPackage in newMaterials) {
				if (!matPackage.isConverted) continue;

				matPackage.material.SetFloat("_Exposure", jf.val);
			}
		}

		protected void SetTintColor(JSONStorableColor jcolor) {
			foreach (MatPackage matPackage in newMaterials) {
				if (!matPackage.isConverted) continue;
				matPackage.material.SetColor("_Tint", jcolor.colorPicker.currentColor);
			}
		}


	}
}